import { readFileSync } from 'fs';

import { getCollection, type CollectionEntry } from 'astro:content';

import { giteaApi, type Release } from 'gitea-js';
const api = giteaApi('https://codeberg.org', {});

let _releases: Release[];
let _releaseBlogPosts: CollectionEntry<'blog'>[];

const semverRegex = /^v((\d+)\.(\d+)\.(\d+))/;

enum Comparison {
	LeftIsGreater = -1,
	Equal,
	LeftIsLower,
}

function compare(a: number, b: number) {
	if (a < b) {
		return Comparison.LeftIsLower;
	}
	if (a > b) {
		return Comparison.LeftIsGreater;
	}
	return Comparison.Equal;
}

export class ReleaseVersion {
	readonly release?: Release;
	readonly version: string;
	readonly major: number;
	readonly minor: number;
	readonly patch: number;

	constructor(release?: Release) {
		this.release = release;
		const [_, version, major, minor, patch] = semverRegex.exec(release?.tag_name || 'v0.0.0') || [];

		this.version = version;
		this.major = parseInt(major);
		this.minor = parseInt(minor);
		this.patch = parseInt(patch);
	}

	compare(other: ReleaseVersion) {
		const majorCmp = compare(this.major, other.major);
		if (majorCmp != Comparison.Equal) {
			return majorCmp;
		}

		const minorCmp = compare(this.minor, other.minor);
		if (minorCmp != Comparison.Equal) {
			return minorCmp;
		}
		return compare(this.patch, other.patch);
	}

	docsTag() {
		return `v${this.major}.${this.minor}`;
	}
}

const _latestReleases = {
	stable: new ReleaseVersion(),
	lts: new ReleaseVersion(),
};

interface LtsRelease {
	major: string | number;
	minor: string | number;
	cut: string;
	release: string;
	eol: string;
	lts: boolean;
}

/** Fetch a list of the latest Forgejo releases from Codeberg using the Forgejo API. */
export const fetchReleases = async () => {
	await ensureReleasesCached();
	return _releases;
};

async function ensureReleasesCached() {
	if (_releases) {
		return;
	}
	const response = await api.repos.repoListReleases('forgejo', 'forgejo');
	_releases = response.data;

	const releaseScheduleJSON = readFileSync('forgejo-docs/next/release-schedule.json', 'utf8');
	const releaseSchedule = JSON.parse(releaseScheduleJSON);
	const ltsReleases = Object.fromEntries(releaseSchedule.map((x: LtsRelease) => [x.major + '.' + x.minor, x.lts]));

	_releases.forEach((release) => {
		const rv = new ReleaseVersion(release);

		const stableComparison = _latestReleases.stable.compare(rv);
		if (stableComparison == Comparison.LeftIsLower) {
			_latestReleases.stable = rv;
		}

		const majorMinor = rv.major + '.' + rv.minor;
		if (majorMinor in ltsReleases && ltsReleases[majorMinor]) {
			if (_latestReleases.lts.compare(rv) == Comparison.LeftIsLower) {
				_latestReleases.lts = rv;
			}
		}
	});
}

/** Get the latest Forgejo release from Codeberg using the Forgejo API. */
export const getLatestReleaseVersion = async () => {
	await ensureReleasesCached();
	const releaseVersion = _latestReleases.stable;
	if (!releaseVersion?.release) {
		throw new Error('no stable release!');
	}
	return releaseVersion;
};

export const releaseLabel = async (release: Release) => {
	await ensureReleasesCached();
	switch (release.name) {
		case _latestReleases.stable.release?.name:
			return 'Stable';
		case _latestReleases.lts.release?.name:
			return 'LTS';
	}
	return '';
};

export const getReleaseNotesUrl = async (rv: ReleaseVersion) => {
	if (rv.major >= 9 || (rv.major == 8 && rv.patch > 0) || (rv.major == 7 && rv.patch > 6)) {
		return `https://codeberg.org/forgejo/forgejo/src/branch/forgejo/release-notes-published/${rv.version}.md`;
	} else {
		const versionSlug = rv.version.replaceAll('.', '-');
		return `https://codeberg.org/forgejo/forgejo/src/branch/forgejo/RELEASE-NOTES.md#${versionSlug}`;
	}
};

/** Get the blog post for a release, if there is one. */
export const getReleaseBlogPost = async (rv: ReleaseVersion) => {
	if (!_releaseBlogPosts) {
		_releaseBlogPosts = await getCollection('blog', ({ data }) => data.release);
	}
	return _releaseBlogPosts.find(({ data }) => data.release == rv.version);
};
